module Machine.Run where

import Machine.Type
import Machine.World
import Machine.Wahrheit

import Array
import ToDoc
import Set
import FiniteMap
import Maybe

run :: ( Ask state quest, Act state act, Initial obj state )
    => Program quest act -> Int -> obj -> ( Wahrheit, Doc )
run p i o = execute $ runtime p i $ initial o

data Runtime q a s =
     Runtime { program :: Array Int ( Statement q a )
	     , state   :: s   -- external state
	     , pc      :: Int -- next instruction
	     , step    :: Int -- number of instruction so far
	     , max_steps :: Int
	     , labels  :: FiniteMap String Int
	     , values  :: FiniteMap String String
	     , traced_labels :: Set String
	     , watched_values :: Set String
	     }

runtime :: Program q a -> Int -> s -> Runtime q a s
runtime p i s = r 
  where r = Runtime 
	    { program = listArray (1, length $ statements p) $ statements p
	    , state = s
	    , pc = fromMaybe 1 $ lookupFM (labels r) "start"
	    , step = 0
	    , max_steps = i
	    , labels = listToFM $ do
	          ( i, Label l ) <- assocs $ program r
		  return (l, i)
	    , values = emptyFM
	    , traced_labels = trace p
	    , watched_values = watch p
	    }

inform :: ( Ask s q, Act s a ) => Runtime q a s -> Doc
inform r =
    fsep $ punctuate comma
	 [ text "step:" <+> toDoc ( step r )
	 , text "state:" <+> toDoc ( state r )
	 , text "mem:" <+> toDoc ( values r )
	 , text "pc:" <+> toDoc ( pc r )
	 , text "next:" <+> toDoc ( program r ! pc r )
	 ]

execute :: ( Ask s q, Act s a ) => Runtime q a s -> (Wahrheit, Doc)
execute r =
    let 
	inst =  program r ! pc r
	i = inform r
	informed ( f, d ) t = ( f, t <+> i $$ d )
    in 
       if step r >= max_steps r 
       then informed ( Vielleicht, empty ) $ text "timeout"
       else case inst of
         Label l -> 
	     let
		 track x = if l `elementOf` ( traced_labels r )
			   then informed x (text "TRACE") else x
	     in	 track $ next r

	 Accept -> informed ( Ja, empty ) $ text "done"
	 Reject -> informed ( Nein, empty ) $ text "done"

	 Goto lab -> goto r lab
	 If (x, y) lab -> 
	    if evaluate r x == evaluate r y
	    then goto r lab else next r

	 Assign loc exp -> 
	     let val = evaluate r exp
		 look x = if loc `elementOf` watched_values r 
			  then informed x (text "WATCH") else x
	     in	 look $ next $ r { values = addToFM (values r) loc val }

	 Do action -> next $ r { state = act (state r) action }


evaluate :: ( Ask s q, Act s a ) => Runtime q a s -> Value q -> String
evaluate r ( Const s ) = s
evaluate r ( Ref loc ) = fromMaybe "" $ lookupFM (values r) $ loc
evaluate r ( Ask q   ) = ask ( state r ) q 


goto :: ( Ask s q, Act s a ) => Runtime q a s -> String -> ( Wahrheit, Doc )
goto r lab = 
    execute $ r { pc = fromMaybe 0 $ lookupFM ( labels r ) lab
		, step = succ $ step r
		}

next :: ( Ask s q, Act s a ) => Runtime q a s -> ( Wahrheit, Doc )
next r = 
    execute $ r { pc = succ $ pc r
		, step = succ $ step r
		}
